#
# Module: __init__.py
#
# Author: Leon Martin
# Copyright © 2024 Charité Universitätsmedizin Berlin.
# Licensed under the EUPL-1.2-or-later
#
"""
TVB-O Parse
=====

This module provides functions for parsing different types of data to extract model components and metadata.
"""
